/*
 * Bilinear transformation routines.
 *
 * Implements pole/zero mapping between s-plane and z-plane.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "macros.h"
#include "common.h"
#include "complex.h"
#include "util.h"

double
bilinear_frequency_digital(double freq, double sample_rate)
{
	return tan(0.5 * freq * sample_rate);
}

/*
 * Warps frequency to z-plane frequency and back to analog.
 *
 * freq:		warped frequency
 * sample_rate:		sampling rate (1/F)
 *
 * Returns: warped frequency.
 *
 * Uses the following formula:
 *
 * f' = 2/T*arctg(f/2)
 */

double
bilinear_frequency_warp(double freq, double sample_rate)
{
	double digital, analog;

	digital = atan(0.5 * freq * sample_rate);
	analog = (2.0 / sample_rate) * tan(digital);

	return analog;
}

/*
 * bilinear_transform - map poles/zeros of s-plane to z-plane
 *
 * splane:		s-plane transfer function
 * sample_rate:		sampling rate (1/F_sample)
 * calculate_hzero:	0 - do no calculate constant gain, 1 - calculate constant gain
 *
 * Returns: z-plane transfer function.
 *
 * This implementation uses the following formula:
 *
 *      2 + sT
 *  z = ------
 *      2 - sT
 */

struct transfer_func *
bilinear_transform(struct transfer_func *splane, double sample_rate, int calculate_hzero)
{
	struct transfer_func *zplane;
	complex_t two, num, den;
	double h;
	int i;

	OBJ_NEW(zplane);

	two = c_num(2.0, 0.0);

	zplane->h_zero = splane->h_zero;

	/* Map splane zeros to zplane using BLT. */
	if (splane->nz > 0) {
		zplane->zeros = xmalloc0(splane->nz * sizeof(*zplane->zeros));
		zplane->nz = splane->nz;

		for (i = 0; i < splane->nz; i++) {
			/* Zero s=0 maps to 1 at zplane circle.*/
			if ((splane->zeros[i].re == 0.0) && (splane->zeros[i].im == 0.0))
				zplane->zeros[i] = c_num(1.0, 0.0);
			else if ((splane->zeros[i].re == 0.0) && (splane->zeros[i].im == INFINITY))
				zplane->zeros[i] = c_num(-1.0, 0.0);
			else {
				num = c_sum(two, c_mult(splane->zeros[i], c_num(sample_rate, 0.0)));
				den = c_sub(two, c_mult(splane->zeros[i], c_num(sample_rate, 0.0)));
				zplane->zeros[i] = c_div(num, den);
			}
		}
	} 
	
	/* Map splane poles to zplane using BLT. */
	if (splane->np > 0) {
		zplane->poles = xmalloc0(splane->np * sizeof(*zplane->poles));
		zplane->np = splane->np;

		for (i = 0; i < splane->np; i++) {
			num = c_sum(two, c_mult(c_num(sample_rate, 0.0), splane->poles[i]));
			den = c_sub(two, c_mult(c_num(sample_rate, 0.0), splane->poles[i]));
			zplane->poles[i] = c_div(num, den);
		}
	}

	/* Calculate constant gain. */
	if (calculate_hzero) {	
		h = 1.0;
		den = c_num(1.0, 0.0);

		for (i = 0; i < splane->np; i++) {
			h *= sample_rate; /* sample_rate^N */
			den = c_mult(den, c_sub(two, c_mult(splane->poles[i], c_num(sample_rate, 0.0))));
		}

		zplane->h_zero = (splane->h_zero * h) / den.re;
	} else
		zplane->h_zero = splane->h_zero;

	return zplane;
}

/*
 * bilinear_transform_v2 - map poles/zeros of s-plane to z-plane
 *
 * splane:		s-plane transfer function
 * sample_rate:		sampling rate (1/F_sample)
 * calculate_hzero:	0 - do no calculate constant gain, 1 - calculate constant gain
 *
 * Returns: z-plane transfer function.
 *
 * This implementation uses the following formula (without 2/T factor):
 *
 *      1 + s
 *  z = ------
 *      1 - s
 */

struct transfer_func *
bilinear_transform_v2(struct transfer_func *splane, int calculate_hzero)
{
	struct transfer_func *zplane;
	complex_t one, num, den;
	double h;
	int i, n;

	OBJ_NEW(zplane);

	one = c_num(1.0, 0.0);

	zplane->h_zero = splane->h_zero;

	/* Map splane zeros to zplane using BLT. */
	if (splane->nz > 0) {
		zplane->zeros = xmalloc0(splane->nz * sizeof(*zplane->zeros));
		zplane->nz = splane->nz;

		for (i = 0; i < splane->nz; i++) {
			/* Zero s=0 maps to 1 at zplane circle.*/
			if ((splane->zeros[i].re == 0.0) && (splane->zeros[i].im == 0.0))
				zplane->zeros[i] = c_num(1.0, 0.0);
			else if ((splane->zeros[i].re == 0.0) && (splane->zeros[i].im == INFINITY))
               			zplane->zeros[i] = c_num(-1.0, 0.0);
			else {
				num = c_sum(one, splane->zeros[i]);
				den = c_sub(one, splane->zeros[i]);
				zplane->zeros[i] = c_div(num, den);
			}
		}
	}
    
	/* Map splane poles to zplane using BLT. */
	if (splane->np > 0) {
		zplane->poles = xmalloc0(splane->np * sizeof(*zplane->poles));
		zplane->np = splane->np;

		for (i = 0; i < splane->np; i++) {
			num = c_sum(one, splane->poles[i]);
			den = c_sub(one, splane->poles[i]);
			zplane->poles[i] = c_div(num, den);
		}
	}

	/* Calculate constant gain. */
	if (calculate_hzero) {
		h   = 1.0;
		den = c_num(1.0, 0.0);

		for (i = 0; i < splane->np; i++)
			den = c_mult(den, c_sub(one, splane->poles[i]));

		zplane->h_zero = (splane->h_zero / den.re);
	} else
		zplane->h_zero = splane->h_zero;
	
	zplane->h_zero=1.0;
	return zplane;
}
